#include <vector>
#include <iostream>

#include <GL/glut.h>

#include <cmath>

#include "Vec3f.h"
#include "Mat4f.h"
#include "OpenGLMatrixTools.h"

// GLOBALS //

typedef Vec3f Triangle[3];
Triangle t1, t2;

Vec3f point; // add to see
float radius = 0.5;
int numSegments = 15;

// Saved Matrices (missing view)
Mat4f modelMatrix;
float thetaYRotate = 0;
float thetaXRotate = 0;
float mousePreviousY;
float mousePreviousX;

int W_WIDTH = 800;
int W_HEIGHT = 800;

void reloadModelTransform()
{
	modelMatrix = RotateAboutXMatrix( thetaXRotate );
	modelMatrix = RotateAboutYMatrix( thetaYRotate ) * modelMatrix;

	// to rotate around both y and x axis, implement matrix multiplication
	// and multiply matrices together to compose a transform that does both
	// (MRotY * MRotX)
}

Vec3f transformPoint( const Vec3f & v3 )
{
	float col[4], res[4];
	col[0] = v3[0]; 
	col[1] = v3[1]; 
	col[2] = v3[2]; 
	col[3] = 1; //Because we are working with a point, we set w=1

	float coord;
	for (int i=0; i<4; ++i)
	{
		coord = 0.0;
		for( int j=0; j<4; ++j )
		{
			coord += modelMatrix( i, j )*col[j];
		}
		res[i] = coord;
	}

	//Divide by W
	Vec3f ret = Vec3f( res[0], res[1], res[2] );
	ret = ret * (1.0 / res[3]); 

	return ret;
}


void glVertex( Vec3f vec )
{
	vec = transformPoint( vec );

	glVertex3f( vec.x(), vec.y(), vec.z() );
	// or equivalent glVertex3f( vec[0], vec[1], vec[2] );
}

void renderCircle()
{
	//Draw a circle
	glBegin(GL_TRIANGLE_FAN);
	glColor3f(0.2, 1.0, 0.5);	
	glVertex( point );
	float stepSize = 2.0 * M_PI / (numSegments - 1);
	for (int i=0; i< numSegments; ++i)
	{
		Vec3f circleVec = Vec3f(cosf(i*stepSize), sinf(i*stepSize), 0.0);
		Vec3f pointOnCircle = point + circleVec * radius;
		glVertex(pointOnCircle);
	}
	glEnd();
}

void renderGrid()
{
	float gridSpacing = 0.15;
	Vec3f minExtent(-0.8, 0.0, -0.8);
	Vec3f maxExtent(0.8, 0.0, 0.8);

	int numStepsX = (maxExtent.x() - minExtent.x()) / gridSpacing;
	int numStepsZ = (maxExtent.z() - minExtent.z()) / gridSpacing;

	glColor3f(0.7,0.7,0.1);
	glLineWidth(2.0);
	glBegin(GL_LINES);
	for (int i=0; i<numStepsX; ++i)
	{
		//Draw lines in the Z direction
		float u = i / (float) (numStepsX - 1);
		Vec3f p = Vec3f(minExtent.x() * u + maxExtent.x() * (1.0 - u), 0, minExtent.z());
		Vec3f q = Vec3f(minExtent.x() * u + maxExtent.x() * (1.0 - u), 0, maxExtent.z());

		glVertex(p);
		glVertex(q);
	}
	for (int i=0; i<numStepsZ; ++i)
	{
		//Draw lines in the X direction
		float u = i / (float) (numStepsZ - 1);
		Vec3f p = Vec3f(minExtent.x(), 0, minExtent.z() * u + maxExtent.z() * (1.0 - u));
		Vec3f q = Vec3f(maxExtent.z(), 0, minExtent.z() * u + maxExtent.z() * (1.0 - u));

		glVertex(p);
		glVertex(q);
	}	
	glEnd();
}

void renderAxis()
{
	//Draw the axis
	glLineWidth(3.0);
	glBegin(GL_LINES);

	//X
	Vec3f origin(0,0,0);
	Vec3f X(1,0,0), Y(0,1,0), Z(0,0,1);
	glColor3f(1.0,0.0,0.0);
	glVertex(origin);
	glVertex(X);

	//Y
	glColor3f(0.0,1.0,0.0);
	glVertex(origin);
	glVertex(Y);

	//Z
	glColor3f(0.0,0.0,1.0);
	glVertex(origin);
	glVertex(Z);

	glEnd();
}

void renderShapes()
{
	renderAxis();
	renderGrid();
	renderCircle();
}

void display()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	

	renderShapes();

	glutSwapBuffers();
}

void idle()
{
	static float x = 0.0;
	x += 0.05;
	point.z( std::sin(x) * 0.8 );

	glutPostRedisplay();

}

Vec3f screenToWorld( int x, int y )
{
	Vec3f world;
	
	// division by ints issue so cast
	world.x( 2.0 * (float(x)/W_WIDTH) -1.0 );
	world.y( 2.0 * (1.0f - float(y)/W_HEIGHT) -1.0 );
	world[2] = 0.0; // world.z( 0 );

	return world;
}

void mouseButtonFunc( int button, int state, int x, int y )
{
	if( state == GLUT_LEFT_BUTTON )
	{
		mousePreviousY = y;
		mousePreviousX = x;
	}
	
	// Add after
	glutPostRedisplay();
}

void mouseMotionFunc( int x, int y )
{	
	//
	thetaXRotate -= y - mousePreviousY;
	thetaYRotate += x - mousePreviousX;
	mousePreviousY = y;
	mousePreviousX = x;

	reloadModelTransform();
	//	
	
	glutPostRedisplay();
}

void keyboardFunc( unsigned char key, int x, int y )
{
	
}

void resizeFunc( int width, int height )
{
	W_WIDTH = width;
	W_HEIGHT = height;

	glViewport( 0, 0, W_WIDTH, W_HEIGHT );

	glutPostRedisplay();
}

void init()
{
	glEnable(GL_DEPTH_TEST);

	// Setup Triangles
	t1[0] = Vec3f( 0, 1, 0.1 );
	t1[1] = Vec3f( -1.0, 0, 0.1 );
	t1[2] = Vec3f( 0, 0, 0.1 );

	t2[0] = Vec3f( 0, -0.5, 0 );
	t2[1] = t2[0] + Vec3f( 0, 1, 0.5 );
	t2[2] = t2[0] + Vec3f( 0, 1, 0 );

	reloadModelTransform();

	std::cout << "Called in init() " << t2[0] << " " << t2[1] << " " << t2[2] << std::endl;
}

int main( int argc, char** argv )
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);

	// not required by window system to work
	// http://www.opengl.org/documentation/specs/glut/spec3/node11.html
	glutInitWindowSize( W_WIDTH, W_HEIGHT );
	glutInitWindowPosition(0,0);

	glutCreateWindow("CPSC 453 Lab 4");

	glutDisplayFunc( display );
	//NEW
	glutIdleFunc( idle );

	glutMouseFunc( mouseButtonFunc );
	glutMotionFunc( mouseMotionFunc );
	glutKeyboardFunc( keyboardFunc );
	glutReshapeFunc( resizeFunc );

	init();

	glutMainLoop();
	return 0;
}
